<?php

namespace Drupal\commerce_stripe_payment_request_button\Controller;

use Drupal;
use Exception;
use Stripe\Stripe;
use Stripe\PaymentIntent;
use Stripe\Exception\ApiErrorException;

use Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\commerce_stripe\ErrorHelper;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Routing\RouteMatchInterface;

class PaymentRequestButtonCheckout extends ControllerBase {

    /**
     * The entity type manager.
     *
     * @var \Drupal\Core\Entity\EntityTypeManagerInterface
     */
    protected $entityTypeManager;

    /**
     * The messenger.
     *
     * @var \Drupal\Core\Messenger\MessengerInterface
     */
    protected $messenger;

    /**
     * The messenger.
     *
     * @var \Drupal\Core\Routing\RouteMatchInterface
     */
    protected $route_match;

    /**
     * Stripe payment request button helper service.
     *
     * @var \Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface
     */
    protected $paymentRequestButtonService;

    /**
     * Creates a new ViewModeAccessCheck.
     *
     * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
     *   The entity type manager.
     */
    public function __construct(
        EntityTypeManagerInterface $entity_type_manager,
        MessengerInterface $messenger,
        RouteMatchInterface $route_match,
        PaymentRequestButtonHelperServiceInterface $paymentRequestButtonService
    ) {
        $this->entityTypeManager = $entity_type_manager;
        $this->messenger = $messenger;
        $this->routeMatchInterface = $route_match;
        $this->paymentRequestButtonService = $paymentRequestButtonService;
    }

    /**
     * {@inheritdoc}
     */
    public static function create(ContainerInterface $container) {
        return new static(
            $container->get('entity_type.manager'),
            $container->get('messenger'),
            $container->get('current_route_match'),
            $container->get('commerce_stripe_payment_request_button.helper')
        );
    }

    /**
     * Handles the incoming order from the "classic" checkout process
     *
     * @param OrderInterface $commerce_order
     * @param Request $request
     * @return JsonResponse
     */
    public function handleOrder(OrderInterface $commerce_order, Request $request) {
        // Transform JSON values to array
        $values = Json::decode($request->getContent());
        $shipping_info = '';
        if (isset($values["data"]["result"]["shippingOption"]["shippingAddress"])) { //if multiple shipping addresses are available
            $shipping_info = $values["data"]["result"]["shippingOption"]["shippingAddress"];
        } elseif (isset($values["data"]["result"]["shippingAddress"])) {
            $shipping_info = $values["data"]["result"]["shippingAddress"];
        }
        $order = $commerce_order;
        //get the current currency order
        $currency = strtolower($order->getTotalPrice()->getCurrencyCode());

        // Get the configuration of the payment gateway
        $orderPaymentConfiguration = $this->paymentRequestButtonService->getOrderPaymentConfiguration($currency);

        $configName = $orderPaymentConfiguration->configName;
        $config = $orderPaymentConfiguration->config;
        Stripe::setApiKey($config['secret_key']);

        // If order has problems, take it back to cart
        if (empty($order) || $order->getState()->value == 'canceled') {
            $response = [
                'redirect_uri' => Url::fromRoute(Url::fromRoute('commerce_cart.page')
                    ->toString()),
            ];
            return new JsonResponse($response);
        }

        // Check if we already have payment intent
        $intentId = $order->getData('stripe_intent');
        if ($intentId !== NULL) {
            try {
                $intent = $this->paymentRequestButtonService->updateDrupalPaymentByPaymentIntent($order, $intentId);
            } catch (ApiErrorException $e) {
                ErrorHelper::handleException($e);
            }
        } else {
            try {
                $paymentIntentData = [
                    'payment_method' => $values["data"]["payment_method_remote_id"],
                    'amount' => $values["data"]["payment_request"]["total"]["amount"],
                    'currency' => $values["data"]["payment_request"]["currency"],
                    'payment_method_types' => ['card'],
                    'capture_method' => $config['capture_method'],
                    'description' => 'Order ' . $order->id(),
                    'metadata' => ['order_id' => $order->id()],
                    'confirmation_method' => 'automatic',
                    'confirm' => TRUE
                ];
                // payment_request_button_shipping
                // Create payment intent
                $intent = PaymentIntent::create($paymentIntentData);

                if ($order->hasItems()) {
                    // Put order in proper status and assign value
                    $shipping = '';
                    if (isset($values['data']['result']['shippingOption']['id'])) {
                        $shipping = $values['data']['result']['shippingOption']['id'];
                    }
                    $this->paymentRequestButtonService->setOrderData($order, $intent, $values["data"]["result"]["payerEmail"], $configName, $shipping);

                    // Create payment for this order and give it values
                    $payment_gateway = $this->entityTypeManager
                        ->getStorage('commerce_payment_gateway')
                        ->load($configName);

                    $current_user = $order->getCustomerId();

                    $payment = $this->paymentRequestButtonService->createDrupalPayment($order, $payment_gateway, $intent, $current_user);

                    if ($config['capture_method'] == "manual") {
                        $payment->setState("authorization");
                    }
                    $payment->save();

                    $this->paymentRequestButtonService->processProfile($order, $intent->charges['data'][0], $config['request_phone']);

                    // Deal with billing profile


                    // Deal with shipping stuff only if we selected that we want to fetch shipping address
                    if ($config['request_shipping'] && !empty($shipping_info)) {
                        $this->paymentRequestButtonService->handleShipping($order, $shipping_info, $values, $config['request_phone']);
                    }
                }
            } catch (ApiErrorException $e) {
                // Server will return 500 to the frontend.
                $this->messenger->addError($e->getMessage());
                ErrorHelper::handleException($e);
            }
        }

        if ($intent->status == 'requires_confirmation' || $intent->status == 'requires_source_action' || $intent->status == 'requires_action') {
            # Tell the client to handle the action
            $response = [
                'requires_action' => TRUE,
                'payment_intent_client_secret' => $intent->client_secret,
                'payment_method_remote_id' => $values["data"]["payment_method_remote_id"],
                'redirect_uri' => Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id()])
                    ->toString() . '/payment',
            ];
        } else {
            if ($intent->status == 'succeeded' || $intent->status == 'requires_capture') {
                # The payment didn’t need any additional actions and completed!
                # Handle post-payment fulfillment
                $response = [
                    'success' => TRUE,
                    'redirect_uri' => Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id(),])
                        ->toString() . '/payment',
                ];
            } else {
                # Invalid status
                //http_response_code(500);
                $response = ['error' => 'Invalid PaymentIntent status'];
            }
        }
        return new JsonResponse($response);
    }


    //check if order marked as paid
    //check if payment intent paid

    public function handleDecoupledOrder($commerceOrderUuId, $payment_intent) {
        $orders = $this->entityTypeManager
            ->getStorage('commerce_order')
            ->loadByProperties(['uuid' => $commerceOrderUuId]);
        $order = reset($orders);
        $order->setData('stripe_intent', $payment_intent);
        $order->save();
        $payment_gateway = $order->get('payment_gateway')->entity;
        $payment_gateway_plugin = $payment_gateway !== NULL ? $payment_gateway->getPlugin() : NULL;

        return $payment_gateway_plugin->processOrder($order);
    }

    public function updateOrderShipping($commerceOrderUuId, $shippingId) {
        $orders = $this->entityTypeManager->getStorage('commerce_order')
            ->loadByProperties(['uuid' => $commerceOrderUuId]);
        $order = reset($orders);
        try {
            $order->setData('payment_request_button_shipping', $shippingId);
            $order->save();
            return new JsonResponse(['success' => true]);
        } catch (Exception $e) {
            $error_message = $e->getMessage();
            $this->messenger->addError($error_message, TRUE);
            Drupal::logger('commerce_stripe_payment_request_button')
                ->error(t('Failed to update order @order_id shippiping . @message', [
                    '@order_id' => $order->id(),
                    '@message' => $e,
                ]));
        }
        return new JsonResponse(['success' => false]);
    }

    /**
     * Checks access for the form page.
     *
     * @return \Drupal\Core\Access\AccessResult
     *   The access result.
     */
    public function checkAccess() {
        $orderCollection = $this->entityTypeManager->getStorage('commerce_order')
            ->loadByProperties(['uuid' => $this->routeMatchInterface->getParameter('commerceOrderUuId')]);
        $order = reset($orderCollection);
        if (!$order || ($order && $order->getState()->getId() == 'canceled')) {
            return AccessResult::forbidden()->addCacheableDependency($order);
        }

        $access = AccessResult::allowedIf($order->hasItems())
            ->addCacheableDependency($order);
        return $access;
    }
}
